#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <ncurses.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <glob.h>
#include "sel.h"

// Global variables --------------------------------------------------------

file_struct *file_list;
int  file_count = -1;        // How many files found
int  len_page;               // How many lines to display information (0 - n)
int  lines_page;             // How many lines your terminal has got (0 - n)
int  pointer;                // The actually marked filename
int  pointer_pos;            // Absolute Pointer position
int  start_dsp;              // beginning of display of file_list
int  end_dsp;                // end of display of file_list
int  color_dsp;              // color yes=1 no=0
int  search_start;           // pointer-position to start search at
int  search_success;         // last search ok=1 failed=0
char working_dir[len_path];  // current directory
char last_dir[len_path];     // dir before last cd, used to jump back
char home_dir[len_path];     // users home
char start_dir[len_path];    // Start directory (when sel was started)
char search_term[len_line];  // searchterm, used by /

// Settings ----------------------------------------------------------------

char  set_cmd[len_cmd];
char  set_path[len_path];
char  set_mask[len_path];
char  set_type[len_types];
char  set_banner[len_line];
int   set_ask_before_cmd;
int   set_blank_before_cmd;
int   set_use_mask;
int   set_use_full_path;
int   set_long_listing;
int   set_pause_after_cmd;
int   set_quit_after_cmd;
int   set_read_after_cmd;
short set_color_status_fg;
short set_color_status_bg;
short set_color_pointer_fg;
short set_color_pointer_bg;
short set_color_backgnd_fg;
short set_color_backgnd_bg;
short set_color_msg_fg;
short set_color_msg_bg;

// Signal + Error handling -------------------------------------------------

void sig_int(int sig) {
    signal(SIGINT, SIG_IGN);
    sel_exit(error_noaction);
}

void sig_term(int sig) {
    signal(SIGTERM, SIG_IGN);
    sel_exit(error_noaction);
}

void sig_winch(int sig) {
    signal(SIGWINCH, SIG_IGN);
    error("Can't change window-size while running sel", error_terminal);
}

void sig_init() {
    if (signal(SIGINT, sig_int) == SIG_ERR) {
        printf("Error: can't handle SIGINT\n");
        exit(error_signal);
    }
    if (signal(SIGTERM, sig_term) == SIG_ERR) {
        printf("Error: can't handle SIGTERM\n");
        exit(error_signal);
    }
    if (signal(SIGWINCH, sig_winch) == SIG_ERR) {
        printf("Error: can't handle SIGWINCH\n");
        exit(error_signal);
    }
}

void error(char *errtxt, int rc) {
    if (curscr != NULL) endwin();
    printf("%s\n", errtxt);
    exit(rc);
}

// the normal way to leave sel

void sel_exit(int rc) {
    if (curscr != NULL) {
        endwin();
        printf(lf);
    }
    exit(rc);
}

// User interaction --------------------------------------------------------

void help(void) {
    printf(version);
    printf(lf);
    printf(lf);
    printf("Syntax: select [options] [files]\n");
    printf(lf);
    printf("Options\n");
    printf(lf);
    printf("-a Ask before executing a command\n");
    printf("-b Text to display as headline\n");
    printf("-c Command to execute. 'percent' will be replaced by selected entry\n");
    printf("-d Start directory\n");
    printf("-f Use complete (absolute) filename with full path in executed commands\n");
    printf("-h Show this help\n");
    printf("-l Long file-listing with uid, gid, size, date and permissions\n");
    printf("-m filemask\n");
    printf("-w Wait after executing a command\n");
    printf("-q Quit after executing a command\n");
    printf("-u Read directory after executing a command\n");
    printf("-s Blank screen before executing a command\n");
    printf("-t File-type filter-string. Possible values: f,l,d,c,b,p,s (file, link,\n");
    printf("   directory, char-dev., block-dev., fifo, socket) Combinations such as\n");
    printf("   -t fdl are allowed\n");
    printf(lf);
    printf("Color setting options\n");
    printf(lf);
    printf("--color-status_fg color-value  Set foreground-color of statusline\n");
    printf("--color-status_bg color-value  Set background-color of statusline\n");
    printf("--color-pointer-fg color-value Set foreground-color of pointer\n");
    printf("--color-pointer-bg color-value Set background-color of pointer\n");
    printf("--color-backgnd-fg color-value Set background-color of file-list\n");
    printf("--color-backgnd-bg color-value Set foreground-color of file-list\n");
    printf("--color-msg-fg color-value     Set foreground-color of warnings\n");
    printf("--color-msg-bg color-value     Set background-color of warnings\n");
    printf(lf);
    printf("Availiable color-values are:\n");
    printf(lf);
    printf("black\n");
    printf("red\n");
    printf("green\n");
    printf("yellow\n");
    printf("blue\n");
    printf("magenta\n");
    printf("cyan\n");
    printf("white\n");
    printf(lf);
    printf("[files] can be wildcards such as *.c; if you don't specify [files], sel\n");
    printf("displays all files (*).\n");
    sel_exit(error_ok);
}

void help_online(void) {
    endwin();
    if (debug) system("less help.txt");
    else system("less /usr/local/share/sel/help.txt");
}
    
void status_put(char instr[]) {
    attron(A_BOLD);
    if (color_dsp) attron(COLOR_PAIR(COLOR_STATUS));
    else attron(A_REVERSE);
    mvprintw(lines_page + 1,0,"%-79s", instr);
    attroff(A_BOLD);
    if (color_dsp) attroff(COLOR_PAIR(COLOR_STATUS));
    else attroff(A_REVERSE);
}

// Initializations ---------------------------------------------------------

int term_getlines(int fd) {
    struct winsize size;
    
    if (ioctl(fd, TIOCGWINSZ, (char *)&size) < 0) return 0;
    return size.ws_row;
}

int term_getcols(int fd) {
    struct winsize size;
    
    if (ioctl(fd, TIOCGWINSZ, (char *)&size) < 0) return 0;
    return size.ws_col;
}

void init(void) {
    // initialize signals
    sig_init();
    
    // set defaults
    strcpy(set_banner, version);
    strcpy(set_path, ".");
    strcpy(set_mask, "*");
    strcpy(set_type, "*");
    set_cmd[0]           = '\0';
    set_use_mask         = OFF;
    set_use_full_path    = OFF;
    set_ask_before_cmd   = OFF;
    set_blank_before_cmd = OFF;
    set_pause_after_cmd  = OFF;
    set_quit_after_cmd   = OFF;
    set_read_after_cmd   = OFF;
    set_color_status_fg  = DEF_COLOR_STATUS_FG;
    set_color_status_bg  = DEF_COLOR_STATUS_BG;
    set_color_pointer_fg = DEF_COLOR_POINTER_FG;
    set_color_pointer_bg = DEF_COLOR_POINTER_BG;
    set_color_backgnd_fg = DEF_COLOR_BACKGND_FG;
    set_color_backgnd_bg = DEF_COLOR_BACKGND_BG;
    set_color_msg_fg     = DEF_COLOR_MSG_FG;
    set_color_msg_bg     = DEF_COLOR_MSG_BG;

    search_term[0] = '\0';
    
    // Find out number of lines on your console, xterm or whatever
    lines_page = term_getlines(STDIN_FILENO) - 2;
    if (lines_page < 3) error("Window too small", error_terminal);

    // Check columns
    if (term_getcols(STDIN_FILENO) != len_line)
        error("Wrong nr. of columns: We need 80!", error_terminal);
    
    // Find out home-directory
    strcpy(home_dir, getenv("HOME"));

    // Initialize the file_list-array
    file_list = (file_struct *)malloc(0);
}

void display_init(void) {
    file_count  = -1;
    pointer     =  0;
    pointer_pos =  0;
    start_dsp   =  0;
    len_page    =  lines_page - 1;
    end_dsp     =  len_page;
}

void display_limit(void) {
    if (file_count < len_page) {
        len_page = file_count;
        end_dsp = file_count;
    }
}

// Color functions ---------------------------------------------------------

short color_get (char instr[]) {
    if (strcmp(instr, "black")   == 0) return COLOR_BLACK;
    if (strcmp(instr, "red")     == 0) return COLOR_RED;
    if (strcmp(instr, "green")   == 0) return COLOR_GREEN;
    if (strcmp(instr, "yellow")  == 0) return COLOR_YELLOW;
    if (strcmp(instr, "blue")    == 0) return COLOR_BLUE;
    if (strcmp(instr, "magenta") == 0) return COLOR_MAGENTA;
    if (strcmp(instr, "cyan")    == 0) return COLOR_CYAN;
    if (strcmp(instr, "white")   == 0) return COLOR_WHITE;
    return(0);
}

void color_init(void) {
    color_dsp = FALSE;
    if (has_colors()) {
        color_dsp = TRUE;
        start_color();
        init_pair(COLOR_STATUS,  set_color_status_fg,  set_color_status_bg);
        init_pair(COLOR_POINTER, set_color_pointer_fg, set_color_pointer_bg);
        init_pair(COLOR_BACKGND, set_color_backgnd_fg, set_color_backgnd_bg);
        init_pair(COLOR_MSG,     set_color_msg_fg,     set_color_msg_bg);
        bkgdset(COLOR_PAIR(COLOR_BACKGND) | ' ');
        clear();
    }
}

// Helpers for listing_get -------------------------------------------------

void listing_sort(void) {
    int i, j;
    file_struct temp;
    
    for (i = 0; i < file_count; i++) {
        for (j = 0; j < file_count; j++) {
            if (strcmp(file_list[j].name, file_list[j + 1].name) > 0) {
                temp = file_list[j];
                file_list[j] = file_list[j + 1];
                file_list[j + 1] = temp;
            }
        }
    }
}

char *get_uname(int uid) {
    struct passwd *pw_entry;
    static char tmp[len_line];

    pw_entry = getpwuid(uid);
    if (pw_entry != NULL) return pw_entry->pw_name;
    else {
        sprintf(tmp, "%d", uid);
        return tmp;
    }
}

char *get_gname(int gid) {
    struct group *grp_entry;
    static char tmp[len_line];

    grp_entry = getgrgid(gid);
    if (grp_entry != NULL) return grp_entry->gr_name;
    else {
        sprintf(tmp, "%d", gid);
        return tmp;
    }
}

char *timestring(time_t t, char * buf) {
    struct tm * local;
    char tmp[len_line];

    local = localtime(&t);
    strftime(tmp, len_time, "%x ", local);
    strcpy(buf,tmp);
    strftime(tmp, len_time, "%X", local);
    strcat(buf,tmp);
    return buf;
}

// type_file: returns one character representing the type of the file

char type_file(char *filename) {
    struct stat file_stat;
    char type;

    if (stat(filename, &file_stat) == 1) {
        error("Can't stat file", error_intern);
    } else {
        if      (S_ISREG(file_stat.st_mode))  type = 'f'; // regular file
        else if (S_ISDIR(file_stat.st_mode))  type = 'd'; // directory
        else if (S_ISCHR(file_stat.st_mode))  type = 'c'; // char-device
        else if (S_ISBLK(file_stat.st_mode))  type = 'b'; // block-device
        else if (S_ISLNK(file_stat.st_mode))  type = 'l'; // link
        else if (S_ISFIFO(file_stat.st_mode)) type = 'p'; // Named pipe (fifo)
        else if (S_ISSOCK(file_stat.st_mode)) type = 's'; // Socket
        else type = '?';
    }
    return(type);
}

char type_file_l(char *filename) {
    struct stat file_lstat;
    char type;

    if (lstat(filename, &file_lstat) == 1) {
        error("Can't stat file", error_intern);
    } else {
        if      (S_ISREG(file_lstat.st_mode))  type = 'f'; // regular file
        else if (S_ISDIR(file_lstat.st_mode))  type = 'd'; // directory
        else if (S_ISCHR(file_lstat.st_mode))  type = 'c'; // char-device
        else if (S_ISBLK(file_lstat.st_mode))  type = 'b'; // block-device
        else if (S_ISLNK(file_lstat.st_mode))  type = 'l'; // link
        else if (S_ISFIFO(file_lstat.st_mode)) type = 'p'; // Named pipe (fifo)
        else if (S_ISSOCK(file_lstat.st_mode)) type = 's'; // Socket
        else type = '?';
    }
    return(type);
}

// stat_file: Get file-information by stat(), used by get_listing and main

void stat_file(int file_number) {
    struct stat file_lstat;

    if (lstat(file_list[file_number].name, &file_lstat) == 1) {
        error("Can't stat file", error_intern);
    } else {
        strcpy(file_list[file_number].uname, get_uname(file_lstat.st_uid));
        strcpy(file_list[file_number].gname, get_gname(file_lstat.st_gid));
        sprintf(file_list[file_number].size, "%ld", file_lstat.st_size);
        timestring(file_lstat.st_mtime, file_list[file_number].time);
        file_list[file_number].perm[0] = '\0';

        // put together permission-string
        
        if (S_ISREG(file_lstat.st_mode)) strcat(file_list[file_number].perm, "-");
        else if (S_ISDIR(file_lstat.st_mode)) strcat(file_list[file_number].perm, "d");
        else if (S_ISCHR(file_lstat.st_mode)) strcat(file_list[file_number].perm, "c");
        else if (S_ISBLK(file_lstat.st_mode)) strcat(file_list[file_number].perm, "b");
        else if (S_ISLNK(file_lstat.st_mode)) strcat(file_list[file_number].perm, "l");
        else if (S_ISFIFO(file_lstat.st_mode)) strcat(file_list[file_number].perm, "p");
        else if (S_ISSOCK(file_lstat.st_mode)) strcat(file_list[file_number].perm, "s");

        // file-type

        file_list[file_number].type = type_file(file_list[file_number].name);

        // permissions
        
        if (file_lstat.st_mode & S_IRUSR) strcat(file_list[file_number].perm, "r");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_IWUSR) strcat(file_list[file_number].perm, "w");
        else strcat(file_list[file_number].perm, "-");
        if(file_lstat.st_mode & S_ISUID) strcat(file_list[file_number].perm, "s");
        else if (file_lstat.st_mode & S_IXUSR) strcat(file_list[file_number].perm, "x");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_IRGRP) strcat(file_list[file_number].perm, "r");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_IWGRP) strcat(file_list[file_number].perm, "w");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_ISGID) strcat(file_list[file_number].perm, "s");
        else if (file_lstat.st_mode & S_IXGRP) strcat(file_list[file_number].perm, "x");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_IROTH) strcat(file_list[file_number].perm, "r");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_IWOTH) strcat(file_list[file_number].perm, "w");
        else strcat(file_list[file_number].perm, "-");
        if (file_lstat.st_mode & S_IXOTH) strcat(file_list[file_number].perm, "x");
        else strcat(file_list[file_number].perm, "-");
    }
}

// listing_get: read directory ---------------------------------------------

int listing_get(char *path) {
    struct dirent *direntz;
    DIR *dirz;
    int n = -1;

    dirz = opendir(path);
    if (dirz != NULL) {
        while((direntz = readdir(dirz)) != NULL) {
            if (strchr(set_type, '*') == NULL) {
                if (strchr(set_type, type_file_l(direntz->d_name)) == NULL) continue;
            }
            n++;
            file_list = (file_struct *)realloc(file_list, ((sizeof(file_struct)) * (n + 1)));
            strcpy(file_list[n].name,direntz->d_name);
            stat_file(n);
        }
        closedir(dirz);
        file_count = n;
        listing_sort();
        display_limit();
        return(0);
    } else {
        closedir(dirz);
        return(1);
    }
}

// listing_glob: expand wildcards passed to sel with the "-m" argument

int listing_glob(void) {
    glob_t result;
    int flags;
    int i;
    int n = -1;

    flags = 0;
    if (glob(set_mask, flags, NULL, &result) != 0)
        error("no such file or directory", error_cmdline);
    for (i = 0; i < result.gl_pathc; i++) {
        if (strchr(set_type, '*') == NULL) {
            if (strchr(set_type, type_file_l(result.gl_pathv[i])) == NULL) continue;
        }
        n++;
        file_list = (file_struct *)realloc(file_list, ((sizeof(file_struct)) * (n + 1)));
        strcpy(file_list[n].name,result.gl_pathv[i]);
        stat_file(n);
    }
    file_count = n;
    listing_sort();
    display_limit();
    return(0);
}

// listing_put: show a file-listing ----------------------------------------

char* limit_str(char instr[len_line], int maxlen) {
    char *buf;
    
    if ((buf = strdup(instr)) == NULL) error("strdup", error_intern);
    if (strlen(buf) > maxlen) {
        buf[maxlen - 1] = '~';
        buf[maxlen] = '\0';
    }
    return buf;
}

void listing_put(void) {
    char status_line[len_path];
    int i;

    // Display headline

    attron(A_BOLD);
    if (color_dsp) attron(COLOR_PAIR(COLOR_STATUS));
    else attron(A_REVERSE);
    mvprintw(0,0,"%-80s", set_banner);
    attroff(A_BOLD);
    if (color_dsp) attroff(COLOR_PAIR(COLOR_STATUS));
    else attroff(A_REVERSE);

    // Display the file-list (or not...)
    
    if (file_count < 0) {
        move(2,1);
        attron(A_BOLD);
        if (color_dsp) attron(COLOR_PAIR(COLOR_MSG));
        else attron(A_REVERSE);
        printw("No files / no permission. Go back with <d> or <r>");
        attroff(A_BOLD);
        if (color_dsp) attroff(COLOR_PAIR(COLOR_MSG));
        else attroff(A_REVERSE);
    } else {
        move(1,0);
        for(i = start_dsp; i <= end_dsp; i++) {
            if (i == pointer) {
                if (color_dsp) attrset(A_BOLD|COLOR_PAIR(COLOR_POINTER));
                else attron(A_REVERSE);
            } else if (file_list[i].type == 'd') attrset(A_BOLD);
            if (set_long_listing) printw("%-18s %-10s %-10s %-10s %-10s %-10s",
                                         limit_str(file_list[i].name, 17),
                                         file_list[i].perm,
                                         limit_str(file_list[i].uname, 10),
                                         limit_str(file_list[i].gname, 10),
                                         limit_str(file_list[i].size, 10),
                                         file_list[i].time);
            else printw("%-80s", limit_str(file_list[i].name, 79));
            attrset(A_NORMAL);
        }
        if (debug) {
            mvprintw(0,40,"|P:%d PP:%d S:%d E:%d FC:%d LP:%d|",
                     pointer,
                     pointer_pos,
                     start_dsp,
                     end_dsp,
                     file_count,
                     len_page);
        }
    }
    
    // Build and display the status-line

    status_line[0] = '\0';
    strcat(status_line, working_dir);
    if (status_line[strlen(status_line) - 1] != '/') strcat(status_line, "/");
    strcat(status_line, file_list[pointer].name);
    strcpy(status_line, limit_str(status_line, 72));
    sprintf(status_line, "%-72s", status_line);
    strcat(status_line, " h=help");
    status_put(status_line);
    refresh();
}

// Functions concerning command-execution ----------------------------------

int ask_before_cmd(char cmd[]) {
    char input;
    int res;

    res = 0;
    clear();
    move(0,0);
    printw("The following command will be executed:\n\n");
    printw(cmd);
    printw(lf);
    printw(lf);
    printw("Continue? (y/n) ");
    input = getch();
    if (input == 'y') res = 1;
    return(res);
}

void pause_after_cmd(void) {
    char input;
    
    printf("Hit <RETURN> to continue");
    FLUSH;
    input = getch();
    clear();
    return;
}

void execute(void) {
    int i;
    char cmd[len_cmd];
    char filename[len_path];
    char token[2];

    if (set_use_full_path) {
        strcpy(filename, working_dir);
        strcat(filename, "/");
        strcat(filename, file_list[pointer].name);
    } else strcpy(filename, file_list[pointer].name);
    if (strchr(set_cmd, '%') == NULL) {
        strcpy(cmd, "");
        if (strlen(set_cmd) > 0) {
            strcat(cmd, set_cmd);
            strcat(cmd, " ");
        }
        strcat(cmd, filename);
    } else {
        cmd[0] = '\0';
        i = 0;
        while(set_cmd[i] != '\0') {
            if(set_cmd[i] == '%') {
                strcat(cmd, filename);
            } else {
                token[0] = set_cmd[i];
                token[1] = '\0';
                strcat(cmd, token);
            }
            i++;
        }
    }
    if (set_ask_before_cmd) {
        if ((ask_before_cmd(cmd)) == 0) {
            clear();
            return;
        }
    }
    status_put("Running...");
    refresh();
    endwin();
    if (set_blank_before_cmd) system("clear");
    system(cmd);
    if (set_pause_after_cmd) pause_after_cmd();
    if (set_quit_after_cmd) exit(0);

    // Re-read directory

    if (set_read_after_cmd) {
        display_init();
        if (set_use_mask) listing_glob();
        else listing_get(".");
        clear();
        listing_put();
    }
}

void changedir(char dir[]) {
    char *new_dir;

    if ((new_dir = strdup(dir)) == NULL) error("strdup", error_intern);
    if (access(new_dir, X_OK) == 0) {
        strcpy(last_dir, working_dir);
        chdir(new_dir);
        getcwd(working_dir, len_path);
        display_init();
        if (set_use_mask) listing_glob();
        else listing_get(".");
        clear();
    }
}

// scrolling ---------------------------------------------------------------

void scrollup(void) {
    start_dsp--;
    end_dsp--;
}

void scrolldown(void) {
    start_dsp++;
    end_dsp++;
}

// code for online-commands not included in main() --------------------------

int find_next() {
    int i, result;

    result = -1;
    
    echo();
    nocbreak();

    attron(A_BOLD);
    if (color_dsp) attron(COLOR_PAIR(COLOR_STATUS));
    else attron(A_REVERSE);
    mvprintw(lines_page + 1,0,"%s", "Searchterm:");
    clrtoeol();
    attroff(A_BOLD);
    if (color_dsp) attroff(COLOR_PAIR(COLOR_STATUS));
    else attroff(A_REVERSE);
    printw(" ");

    scanw("%80s", search_term);
    noecho();
    cbreak();

    if (search_term[0] != '\0') {
        search_success = 1;
        for (i = search_start; i <= file_count; i++) {
            if (strstr(file_list[i].name, search_term) != NULL) {
                result = i;
                break;
            }
        }
    }
    return(result);
}

// main --------------------------------------------------------------------

int main(int argc, char *argv[]) {
    int input;
    int go = 1;
    int i  = 0;
    int n  = -1;
    int entries = 0;
    int file_input = 0; // 0: Use internal listing_get 1: Files given on the command-line
    int search_result;

    init();

    // Read commandline-arguments
    
    for (i = 1; i <= argc; i++) {
        if (argv[i] != NULL) {
            if (argv[i][0] != '-') { // Filenames, given on the cmd-line
                file_input = 1;
                if (access(argv[i], F_OK) == 0) {
                    if (strchr(set_type, '*') == NULL) {
                        if (strchr(set_type, type_file_l(argv[i])) == NULL) continue;
                    }
                    n++;
                    file_list = (file_struct *)realloc(file_list, ((sizeof(file_struct)) * (n + 1)));
                    strcpy(file_list[n].name,argv[i]);
                    stat_file(n);
                }
            } else {
                if (strcmp(argv[i],"-a") == 0) set_ask_before_cmd = ON;
                else if (strcmp(argv[i],"-b") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        strcpy(set_banner,limit_str(argv[i+1], len_line - 1));
                        i++;
                    } else error("-b without value. Use -b \"My files\"", error_cmdline);
                }
                else if (strcmp(argv[i],"-c") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        strcpy(set_cmd,argv[i+1]);
                        i++;
                    } else error("-c without value. Use -c \"command(s)\"", error_cmdline);
                }
                else if (strcmp(argv[i],"-d") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                            strcpy(set_path,argv[i+1]);
                            i++;
                    } else error("-d without value. Use -d /directory", error_cmdline);
                }
                else if (strcmp(argv[i],"-f") == 0) set_use_full_path = ON;
                else if (strcmp(argv[i],"-h") == 0) help();
                else if (strcmp(argv[i],"-l") == 0) set_long_listing = ON;
                else if (strcmp(argv[i],"-m") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_use_mask = ON;
                        strcpy(set_mask,argv[i+1]);
                        i++;
                    } else error("-m without value. Use -m \"something.*\"", error_cmdline);
                }
                else if (strcmp(argv[i],"-s") == 0) set_blank_before_cmd = ON;
                else if (strcmp(argv[i],"-t") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        strcpy(set_type, argv[i+1]);
                        i++;
                    } else error("-t without value. Use -t [fldcbps]", error_cmdline);
                }
                else if (strcmp(argv[i],"-w") == 0) set_pause_after_cmd = ON;
                else if (strcmp(argv[i],"-q") == 0) set_quit_after_cmd = ON;
                else if (strcmp(argv[i],"-u") == 0) set_read_after_cmd = ON;
                else if (strcmp(argv[i],"--color-status_fg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_status_fg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-status-fg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-status_bg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_status_bg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-status-bg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-pointer-fg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_pointer_fg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-pointer-fg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-pointer-bg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_pointer_bg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-pointer-bg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-backgnd-fg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_backgnd_fg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-backgnd-fg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-backgnd-bg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_backgnd_bg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-backgnd-bg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-msg-fg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_msg_fg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-msg-fg without value", error_cmdline);
                }
                else if (strcmp(argv[i],"--color-msg-bg") == 0) {
                    if ((argv[i+1] != NULL) && (argv[i+1][0] != '-')) {
                        set_color_msg_bg = color_get(argv[i+1]);
                        i++;
                    } else error("--color-msg-bg without value", error_cmdline);
                }
            }
        }
    }

    display_init();
    chdir(set_path);
    getcwd(working_dir, len_path);
    strcpy(start_dir, working_dir);
    strcpy(last_dir, working_dir);

    if (file_input == ON) {
        file_count = n;
        listing_sort();
        display_limit();
    } else {
        if (set_use_mask) {
            listing_glob();
        } else listing_get(".");
    }
    
    initscr();
    keypad(stdscr, 1);
    cbreak();
    noecho();
    color_init();
    listing_put();

    // Input-Loop
    
    while (go) {
        input = getch();
        switch (input) {
        case 'p':
        case KEY_UP: {
            if (pointer > 0) {
                pointer--;
                if (pointer_pos > 0) pointer_pos--;
                else if (pointer_pos == 0)
                    if (start_dsp > 0) scrollup();
            }
            listing_put();
            break;
        }
        case 'n':
        case KEY_DOWN: {
            pointer++;
            if (pointer > file_count) pointer--;
            if (file_count < len_page) {
                if (pointer > file_count) pointer--;
            } else if (file_count >= len_page) {
                if (pointer_pos < len_page) pointer_pos++;
                else if (pointer_pos == len_page) {
                    if (end_dsp < file_count) scrolldown();
                }
            }
            listing_put();
            break;
        }
        case 'P':
        case KEY_PPAGE: {
            if (file_count > len_page) {
                if (start_dsp >= len_page) {
                    start_dsp -= len_page;
                    end_dsp -= len_page;
                    pointer -= len_page;
                } else {
                    entries = start_dsp;
                    start_dsp -= entries;
                    end_dsp -= entries;
                    pointer -= entries;
                }
            } else {
                pointer = 0;
                pointer_pos = 1;
            }
            listing_put();
            break;
        }
        case 'N':
        case KEY_NPAGE: {
            if (file_count > len_page) {
                if ((file_count - end_dsp) >= len_page) {
                    start_dsp += len_page;
                    end_dsp += len_page;
                    pointer += len_page;
                } else {
                    entries = file_count - end_dsp;
                    start_dsp += entries;
                    end_dsp += entries;
                    pointer += entries;
                }
            } else {
                pointer = file_count;
                start_dsp = 0;
                end_dsp = len_page;
            }
            listing_put();
            break;
        }
        case 't': {
            if (file_count <= len_page) {
                pointer = 0;
                pointer_pos = 0;
            } else {
                pointer = 0;
                pointer_pos = 0;
                start_dsp = 0;
                end_dsp = len_page;
            }
            listing_put();
            break;
        }
        case 'b': {
            if (file_count <= len_page) {
                pointer = file_count;
                pointer_pos = file_count;
                start_dsp = 0;
                end_dsp = len_page;
            } else {
                pointer = file_count;
                pointer_pos = len_page;
                start_dsp = file_count - len_page;
                end_dsp = file_count;
            }
            listing_put();
            break;
        }
        case ' ':
        case '\n': {
            if (file_list[pointer].type == 'd') {
                if (input == '\n') changedir(file_list[pointer].name);
                else if (input == ' ') execute();
            }
            else execute();
            listing_put();
            break;
        }
        case KEY_RIGHT: {
            if (file_list[pointer].type == 'd') {
                changedir(file_list[pointer].name);
                listing_put();
            }
            break;
        }
        case KEY_LEFT: {
            if (strcmp(working_dir, "/") != 0) {
                changedir("..");
                listing_put();
            }
            break;
        }
        case '/': {
            if (search_success) search_start = pointer + 1;
            else search_start = 0;
            search_result = find_next();
            if (search_result >= 0) {
                if (file_count <= len_page) {
                    pointer = search_result;
                    pointer_pos = search_result;
                } else {
                    if ((search_result >= start_dsp) && (search_result <= end_dsp)) {
                        pointer = search_result;
                        pointer_pos = pointer - start_dsp;
                    } else {
                        if ((file_count-search_result) >= len_page) {
                            pointer = search_result;
                            pointer_pos = 0;
                            start_dsp = search_result;
                            end_dsp = start_dsp + len_page;
                        } else {
                            pointer = search_result;
                            end_dsp = file_count;
                            start_dsp = end_dsp - len_page;
                            pointer_pos = pointer - start_dsp;
                        }
                    }
                }
                listing_put();
            } else {
                status_put("term not found, next search starts from the top");
                search_success = 0;
            }
            break;
        }
        case 'd': {
            changedir(start_dir);
            listing_put();
            break;
        }
        case 'r': {
            changedir(last_dir);
            listing_put();
            break;
        }
        case 'R': {
            changedir("/");
            listing_put();
            break;
        }
        case 'H':
        case '~': {
            changedir(home_dir);
            listing_put();
            break;
        }
        case 'U': {
            display_init();
            if (set_use_mask) listing_glob();
            else listing_get(".");
            clear();
            listing_put();
            break;
        }
        case 'h': {
            help_online();
            listing_put();
            break;
        }
        case 'a': {
            if (set_ask_before_cmd) {
                set_ask_before_cmd = OFF;
                status_put("ask-before-cmd now off");
            } else {
                set_ask_before_cmd = ON;
                status_put("ask-before-cmd now on");
            }
            break;
        }
        case 'f': {
            if (set_use_full_path) {
                set_use_full_path = OFF;
                status_put("use-full-path now off");
            } else {
                set_use_full_path = ON;
                status_put("use-full-path now on");
            }
            break;
        }
        case 'w': {
            if (set_pause_after_cmd) {
                set_pause_after_cmd = OFF;
                status_put("wait-after-exec now off");
            } else {
                set_pause_after_cmd = ON;
                status_put("wait-after-exec now on");
            }
            break;
        }
        case 'l': {
            if (set_long_listing) {
                set_long_listing = OFF;
                listing_put();
                status_put("long-listing now off");
            } else {
                set_long_listing = ON;
                listing_put();
                status_put("long-listing now on");
            }
            break;
        }
        case 's': {
            if (set_blank_before_cmd) {
                set_blank_before_cmd = OFF;
                status_put("blank-before-exec now off");
            } else {
                set_blank_before_cmd = ON;
                status_put("blank-before-exec now on");
            }
            break;
        }
        case 'u': {
            if(set_read_after_cmd) {
                set_read_after_cmd = OFF;
                status_put("update-after-cmd now off");
            } else {
                set_read_after_cmd = ON;
                status_put("update-after-cmd now on");
            }
            break;
        }
        case 'q': {
            sel_exit(error_noaction);
            break;
        }
        default: break;
        }
    }
    exit(error_ok);
}

// the end
